import {desiredHeight, desiredWidth, getScreen} from "./screen.ts";
import {BG_OUTER} from "./colors.ts";
import {mainFont} from "./font.ts";
import {getInput} from "./input.ts";

type Point = {x: number, y: number}

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position.x = adjust(this.position.x, this.target.x);
    this.position.y = adjust(this.position.y, this.target.y);
  }
}

type GameState = "Overview" | "Gameplay" | "Thralls";

function getScreenLocation(state: GameState): {x: number, y: number} {
  if (state === "Overview") {
    return {x: 0.0, y: 0.0}
  }
  if (state === "Gameplay") {
    return {x: 1.0, y: 0.0}
  }
  if (state === "Thralls") {
    return {x: 0.0, y: 1.0}
  }

  throw `invalid state: ${state}`
}

export class Game {
  msg: string;
  camera: MenuCamera;
  state: GameState;

  constructor() {
    this.msg = "You have been given a gift.";
    this.camera = new MenuCamera({
      position: {x: 0.0, y: 0.0},
      target: {x: 0.0, y: 0.0}
    });
    this.state = "Overview";
  }

  update() {
    if (getInput().isPressed("a") || getInput().isPressed("w")) {
      this.state = "Overview"
    }
    if (getInput().isPressed("d")) {
      this.state = "Gameplay"
    }
    if (getInput().isPressed("s")) {
      this.state = "Thralls"
    }

    if (getInput().isPressed("leftMouse")) {
      this.msg = "Left click, asshole!"
    }
    if (getInput().isReleased("leftMouse")) {
      this.msg = "Un-left click, asshole!"
    }

    this.camera.target = getScreenLocation(this.state);
    this.camera.update();
  }

  draw() {
    let screen = getScreen();
    let ctx = screen.makeContext();

    // draw screen background
    ctx.fillStyle = BG_OUTER;
    ctx.fillRect(0, 0, screen.w, screen.h);

    this.drawOverview();
    // this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let xy = this.getCameraMouseXy();
    if (xy != null) {
      ctx = this.makeCameraContext();
      ctx.fillStyle = "#fff";
      ctx.strokeStyle = "#fff";
      ctx.fillRect(xy.x - 1, xy.y - 1, 3, 3);
    }
  }

  getCameraOffset(): Point {
    let screen = getScreen();
    let {w, h} = screen;
    return {
      x: Math.floor(w * this.camera.position.x),
      y: Math.floor(h * this.camera.position.y)
    };
  }

  getCameraMouseXy(): Point | null {
    let xy = getInput().mouseXy();
    if (xy == null) {
      return null;
    }
    let {x: dx, y: dy} = this.getCameraOffset();
    return {
      x: xy.x + dx,
      y: xy.y + dy,
    };
  }

  makeCameraContext() {
    let screen = getScreen();
    let ctx = screen.makeContext();
    let {x, y} = this.getCameraOffset();

    ctx.translate(-x, -y);
    return ctx;
  }

  getPaneRegionForGameState(gameState: GameState) {
    let screen = getScreen();
    let {w, h} = screen;
    let overallScreenLocation = getScreenLocation(gameState);

    let bigPaneX = overallScreenLocation.x * w;
    let bigPaneY = overallScreenLocation.y * h;
    let bigPaneW = w;
    let bigPaneH = h;

    let smallPaneW = desiredWidth;
    let smallPaneH = desiredHeight;
    let smallPaneX = Math.floor(bigPaneX + (bigPaneW - smallPaneW) / 2)
    let smallPaneY = Math.floor(bigPaneY + (bigPaneH - smallPaneH) / 2)

    return {
      big: {
        position: {x: bigPaneX, y: bigPaneY},
        size: {x: bigPaneW, y: bigPaneH}
      },
      small: {
        position: {x: smallPaneX, y: smallPaneY},
        size: {x: smallPaneW, y: smallPaneH}
      }
    }

  }

  drawOverview() {
    let region = this.getPaneRegionForGameState("Overview")
    let ctx = this.makeCameraContext()
    ctx.translate(region.small.position.x, region.small.position.y)

    // ctx.strokeStyle = "#ffffff";
    // ctx.strokeRect(0.5, 0.5, region.small.size.x - 1, region.small.size.y - 1);

    mainFont.drawText({ctx: ctx, text: this.msg, x: 0, y: 0})
  }
}

export let game = new Game();
